# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from dataclasses import dataclass

from beartype.typing import Union
import jax.numpy as jnp
from jaxtyping import Float
import tensorflow_probability.substrates.jax.bijectors as tfb
import tensorflow_probability.substrates.jax.distributions as tfd

from gpjax.base import param_field
from gpjax.kernels.base import AbstractKernel
from gpjax.kernels.stationary.utils import (
    build_student_t_distribution,
    euclidean_distance,
)
from gpjax.typing import (
    Array,
    ScalarFloat,
)


@dataclass
class Matern12(AbstractKernel):
    r"""The Matérn kernel with smoothness parameter fixed at 0.5."""

    lengthscale: Union[ScalarFloat, Float[Array, " D"]] = param_field(
        jnp.array(1.0), bijector=tfb.Softplus()
    )
    variance: ScalarFloat = param_field(jnp.array(1.0), bijector=tfb.Softplus())
    name: str = "Matérn12"

    def __call__(self, x: Float[Array, " D"], y: Float[Array, " D"]) -> ScalarFloat:
        r"""Compute the Matérn 1/2 kernel between a pair of arrays.

        Evaluate the kernel on a pair of inputs $`(x, y)`$ with
        lengthscale parameter $`\ell`$ and variance $`\sigma^2`$.
        ```math
        k(x, y) = \sigma^2\exp\Bigg(-\frac{\lvert x-y \rvert}{2\ell^2}\Bigg)
        ```

        Args:
            x (Float[Array, " D"]): The left hand argument of the kernel function's call.
            y (Float[Array, " D"]): The right hand argument of the kernel function's call
        Returns:
            ScalarFloat: The value of $`k(x, y)`$
        """
        x = self.slice_input(x) / self.lengthscale
        y = self.slice_input(y) / self.lengthscale
        K = self.variance * jnp.exp(-euclidean_distance(x, y))
        return K.squeeze()

    @property
    def spectral_density(self) -> tfd.Distribution:
        return build_student_t_distribution(nu=1)



# Copyright 2022 The JaxLinOp Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from beartype.typing import (
    Tuple,
    Type,
    Union,
)
import jax
import jax.numpy as jnp
from jaxtyping import Float

from gpjax.linops.linear_operator import LinearOperator
from gpjax.typing import Array


def identity(
    n: int,
) -> "gpjax.linops.identity_linear_operator.IdentityLinearOperator":  # noqa: F821
    """Identity matrix.

    Args:
        n (int): Size of the identity matrix.

    Returns
    -------
        IdentityLinearOperator: Identity matrix of shape [n, n].
    """
    from gpjax.linops.identity_linear_operator import IdentityLinearOperator

    return IdentityLinearOperator(size=n)


def to_dense(obj: Union[Float[Array, "..."], LinearOperator]):
    """
    Ensure an object is a dense matrix.

    Args:
        obj (Union[Float[Array, "..."], LinearOperator]): Linear operator to convert.

    Returns
    -------
        Float[Array, "..."]: Dense matrix.
    """
    if isinstance(obj, jnp.ndarray):
        return obj
    elif isinstance(obj, LinearOperator):
        return obj.to_dense()
    else:
        raise TypeError(
            "object of class {} cannot be made into a Tensor".format(
                obj.__class__.__name__
            )
        )


def to_linear_operator(obj: Union[Float[Array, "..."], LinearOperator]):
    """
    Ensure an object is a linear operator.

    Args:
        obj (Union[Float[Array, "..."], LinearOperator]): Linear operator to convert.

    Returns
    -------
        LinearOperator: Linear operator.
    """
    if isinstance(obj, LinearOperator):
        return obj

    elif isinstance(obj, jnp.ndarray):
        from gpjax.linops.dense_linear_operator import DenseLinearOperator

        return DenseLinearOperator.from_dense(obj)
    else:
        raise TypeError(
            "object of class {} cannot be made into a Tensor".format(
                obj.__class__.__name__
            )
        )


def check_shapes_match(shape1: Tuple[int, ...], shape2: Tuple[int, ...]) -> None:
    """Check shapes of two objects.

    Args:
        shape1 (Tuple[int, "..."]): Shape of the first object.
        shape2 (Tuple[int, "..."]): Shape of the second object.

    Raises
    ------
        ValueError: Shapes of the two objects do not match.
    """
    if shape1 != shape2:
        raise ValueError(
            f"`shape1` must have shape {shape1}, but `shape2` has shape {shape2}."
        )


def default_dtype() -> Union[Type[jnp.float64], Type[jnp.float32]]:
    """Get the default dtype for the linear operator.

    Returns
    -------
        jnp.dtype: Default dtype for the linear operator.
    """
    if jax.config.x64_enabled:
        return jnp.float64
    else:
        return jnp.float32


__all__ = [
    "identity",
    "to_dense",
]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import typing as tp

from jax import vmap
import jax.numpy as jnp
from jaxtyping import Float

from gpjax.kernels.computations import AbstractKernelComputation
from gpjax.linops import (
    ConstantDiagonalLinearOperator,
    DiagonalLinearOperator,
)
from gpjax.typing import Array

Kernel = tp.TypeVar("Kernel", bound="gpjax.kernels.base.AbstractKernel")  # noqa: F821


class ConstantDiagonalKernelComputation(AbstractKernelComputation):
    def gram(
        self, kernel: Kernel, x: Float[Array, "N D"]
    ) -> ConstantDiagonalLinearOperator:
        r"""Compute the Gram matrix.

        Compute Gram covariance operator of the kernel function.

        Args:
            kernel (Kernel): the kernel function.
            x (Float[Array, "N N"]): The inputs to the kernel function.
        """
        value = kernel(x[0], x[0])

        return ConstantDiagonalLinearOperator(
            value=jnp.atleast_1d(value), size=x.shape[0]
        )

    def diagonal(
        self, kernel: Kernel, inputs: Float[Array, "N D"]
    ) -> DiagonalLinearOperator:
        r"""Compute the diagonal Gram matrix's entries.

        For a given kernel, compute the elementwise diagonal of the
        NxN gram matrix on an input matrix of shape $`N\times D`$.

        Args:
            kernel (Kernel): the kernel function.
            inputs (Float[Array, "N D"]): The input matrix.

        Returns
        -------
            DiagonalLinearOperator: The computed diagonal variance entries.
        """
        diag = vmap(lambda x: kernel(x, x))(inputs)

        return DiagonalLinearOperator(diag=diag)

    def cross_covariance(
        self, kernel: Kernel, x: Float[Array, "N D"], y: Float[Array, "M D"]
    ) -> Float[Array, "N M"]:
        r"""Compute the cross-covariance matrix.

        For a given kernel, compute the NxM covariance matrix on a pair of input
        matrices of shape NxD and MxD.

        Args:
            kernel (Kernel): the kernel function.
            x (Float[Array,"N D"]): The input matrix.
            y (Float[Array,"M D"]): The input matrix.

        Returns
        -------
            Float[Array, "N M"]: The computed square Gram matrix.
        """
        # TODO: This is currently a dense implementation. We should implement
        # a sparse LinearOperator for non-square cross-covariance matrices.
        cross_cov = vmap(lambda x: vmap(lambda y: kernel(x, y))(y))(x)
        return cross_cov



# Copyright 2022 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from beartype.typing import (
    Any,
    Optional,
    Tuple,
)
from jax import vmap
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Bool,
    Float,
)
import tensorflow_probability.substrates.jax as tfp

from gpjax.linops import (
    IdentityLinearOperator,
    LinearOperator,
)
from gpjax.typing import (
    Array,
    KeyArray,
    ScalarFloat,
)

tfd = tfp.distributions


def _check_loc_scale(loc: Optional[Any], scale: Optional[Any]) -> None:
    r"""Checks that the inputs are correct."""
    if loc is None and scale is None:
        raise ValueError("At least one of `loc` or `scale` must be specified.")

    if loc is not None and loc.ndim < 1:
        raise ValueError("The parameter `loc` must have at least one dimension.")

    if scale is not None and scale.ndim < 2:
        raise ValueError(
            "The `scale` must have at least two dimensions, but "
            f"`scale.shape = {scale.shape}`."
        )

    if scale is not None and not isinstance(scale, LinearOperator):
        raise ValueError(
            f"scale must be a LinearOperator or a JAX array, but got {type(scale)}"
        )

    if scale is not None and (scale.shape[-1] != scale.shape[-2]):
        raise ValueError(
            f"The `scale` must be a square matrix, but `scale.shape = {scale.shape}`."
        )

    if loc is not None:
        num_dims = loc.shape[-1]
        if scale is not None and (scale.shape[-1] != num_dims):
            raise ValueError(
                f"Shapes are not compatible: `loc.shape = {loc.shape}` and "
                f"`scale.shape = {scale.shape}`."
            )


class GaussianDistribution(tfd.Distribution):
    r"""Multivariate Gaussian distribution with a linear operator scale matrix.

    Args:
        loc (Optional[Float[Array, " N"]]): The mean of the distribution. Defaults to None.
        scale (Optional[LinearOperator]): The scale matrix of the distribution. Defaults to None.

    Returns
    -------
        GaussianDistribution: A multivariate Gaussian distribution with a linear operator scale matrix.
    """

    # TODO: Consider `distrax.transformed.Transformed` object. Can we create a LinearOperator to `distrax.bijector` representation
    # and modify `distrax.MultivariateNormalFromBijector`?

    # TODO: Consider natural and expectation parameterisations in future work.

    def __init__(
        self,
        loc: Optional[Float[Array, " N"]] = None,
        scale: Optional[LinearOperator] = None,
    ) -> None:
        r"""Initialises the distribution."""
        _check_loc_scale(loc, scale)

        # Find dimensionality of the distribution.
        if loc is not None:
            num_dims = loc.shape[-1]

        elif scale is not None:
            num_dims = scale.shape[-1]

        # Set the location to zero vector if unspecified.
        if loc is None:
            loc = jnp.zeros((num_dims,))

        # If not specified, set the scale to the identity matrix.
        if scale is None:
            scale = IdentityLinearOperator(num_dims)

        self.loc = loc
        self.scale = scale

    def mean(self) -> Float[Array, " N"]:
        r"""Calculates the mean."""
        return self.loc

    def median(self) -> Float[Array, " N"]:
        r"""Calculates the median."""
        return self.loc

    def mode(self) -> Float[Array, " N"]:
        r"""Calculates the mode."""
        return self.loc

    def covariance(self) -> Float[Array, "N N"]:
        r"""Calculates the covariance matrix."""
        return self.scale.to_dense()

    def variance(self) -> Float[Array, " N"]:
        r"""Calculates the variance."""
        return self.scale.diagonal()

    def stddev(self) -> Float[Array, " N"]:
        r"""Calculates the standard deviation."""
        return jnp.sqrt(self.scale.diagonal())

    @property
    def event_shape(self) -> Tuple:
        r"""Returns the event shape."""
        return self.loc.shape[-1:]

    def entropy(self) -> ScalarFloat:
        r"""Calculates the entropy of the distribution."""
        return 0.5 * (
            self.event_shape[0] * (1.0 + jnp.log(2.0 * jnp.pi)) + self.scale.log_det()
        )

    def log_prob(
        self, y: Float[Array, " N"], mask: Optional[Bool[Array, " N"]] = None
    ) -> ScalarFloat:
        r"""Calculates the log pdf of the multivariate Gaussian.

        Args:
            y (Optional[Float[Array, " N"]]): the value of which to calculate the log probability.
            mask: (Optional[Bool[Array, " N"]]): the mask for missing values in y.

        Returns
        -------
            ScalarFloat: The log probability of the value.
        """
        mu = self.loc
        sigma = self.scale
        n = mu.shape[-1]
        if mask is not None:
            y = jnp.where(mask, 0.0, y)
            mu = jnp.where(mask, 0.0, mu)
            sigma_masked = jnp.where(mask[None] + mask[:, None], 0.0, sigma.matrix)
            sigma = sigma.replace(
                matrix=jnp.where(jnp.diag(mask), 1 / (2 * jnp.pi), sigma_masked)
            )

        # diff, y - µ
        diff = y - mu

        # compute the pdf, -1/2[ n log(2π) + log|Σ| + (y - µ)ᵀΣ⁻¹(y - µ) ]
        return -0.5 * (
            n * jnp.log(2.0 * jnp.pi) + sigma.log_det() + diff.T @ sigma.solve(diff)
        )

    def _sample_n(self, key: KeyArray, n: int) -> Float[Array, "n N"]:
        r"""Samples from the distribution.

        Args:
            key (KeyArray): The key to use for sampling.

        Returns
        -------
            Float[Array, "n N"]: The samples.
        """
        # Obtain covariance root.
        sqrt = self.scale.to_root()

        # Gather n samples from standard normal distribution Z = [z₁, ..., zₙ]ᵀ.
        Z = jr.normal(key, shape=(n, *self.event_shape))

        # xᵢ ~ N(loc, cov) <=> xᵢ = loc + sqrt zᵢ, where zᵢ ~ N(0, I).
        def affine_transformation(x):
            return self.loc + sqrt @ x

        return vmap(affine_transformation)(Z)

    def sample(
        self, seed: KeyArray, sample_shape: Tuple[int, ...]
    ):  # pylint: disable=useless-super-delegation
        r"""See `Distribution.sample`."""
        return self._sample_n(
            seed, sample_shape[0]
        )  # TODO this looks weird, why ignore the second entry?

    def kl_divergence(self, other: "GaussianDistribution") -> ScalarFloat:
        return _kl_divergence(self, other)


def _check_and_return_dimension(
    q: GaussianDistribution, p: GaussianDistribution
) -> int:
    r"""Checks that the dimensions of the distributions are compatible."""
    if q.event_shape != p.event_shape:
        raise ValueError(
            "Distribution event shapes are not compatible: `q.event_shape ="
            f" {q.event_shape}` and `p.event_shape = {p.event_shape}`. Please check"
            " your mean and covariance shapes."
        )

    return q.event_shape[-1]


def _frobenius_norm_squared(matrix: Float[Array, "N N"]) -> ScalarFloat:
    r"""Calculates the squared Frobenius norm of a matrix."""
    return jnp.sum(jnp.square(matrix))


def _kl_divergence(q: GaussianDistribution, p: GaussianDistribution) -> ScalarFloat:
    r"""KL-divergence between two Gaussians.

    Computes the KL divergence, $`\operatorname{KL}[q\mid\mid p]`$, between two
    multivariate Gaussian distributions $`q(x) = \mathcal{N}(x; \mu_q, \Sigma_q)`$
    and $`p(x) = \mathcal{N}(x; \mu_p, \Sigma_p)`$.

    Args:
        q (GaussianDistribution): A multivariate Gaussian distribution.
        p (GaussianDistribution): A multivariate Gaussian distribution.

    Returns
    -------
        ScalarFloat: The KL divergence between q and p.
    """
    n_dim = _check_and_return_dimension(q, p)

    # Extract q mean and covariance.
    mu_q = q.loc
    sigma_q = q.scale

    # Extract p mean and covariance.
    mu_p = p.loc
    sigma_p = p.scale

    # Find covariance roots.
    sqrt_p = sigma_p.to_root()
    sqrt_q = sigma_q.to_root()

    # diff, μp - μq
    diff = mu_p - mu_q

    # trace term, tr[Σp⁻¹ Σq] = tr[(LpLpᵀ)⁻¹(LqLqᵀ)] = tr[(Lp⁻¹Lq)(Lp⁻¹Lq)ᵀ] = (fr[LqLp⁻¹])²
    trace = _frobenius_norm_squared(
        sqrt_p.solve(sqrt_q.to_dense())
    )  # TODO: Not most efficient, given the `to_dense()` call (e.g., consider diagonal p and q). Need to abstract solving linear operator against another linear operator.

    # Mahalanobis term, (μp - μq)ᵀ Σp⁻¹ (μp - μq) = tr [(μp - μq)ᵀ [LpLpᵀ]⁻¹ (μp - μq)] = (fr[Lp⁻¹(μp - μq)])²
    mahalanobis = jnp.sum(
        jnp.square(sqrt_p.solve(diff))
    )  # TODO: Need to improve this. Perhaps add a Mahalanobis method to ``LinearOperator``s.

    # KL[q(x)||p(x)] = [ [(μp - μq)ᵀ Σp⁻¹ (μp - μq)] - n - log|Σq| + log|Σp| + tr[Σp⁻¹ Σq] ] / 2
    return (mahalanobis - n_dim - sigma_q.log_det() + sigma_p.log_det() + trace) / 2.0


__all__ = [
    "GaussianDistribution",
]



"""Compute Random Fourier Feature (RFF) kernel approximations.  """
from dataclasses import dataclass

from jax.random import PRNGKey
from jaxtyping import Float
import tensorflow_probability.substrates.jax.bijectors as tfb

from gpjax.base import (
    param_field,
    static_field,
)
from gpjax.kernels.base import AbstractKernel
from gpjax.kernels.computations import BasisFunctionComputation
from gpjax.typing import (
    Array,
    KeyArray,
)


@dataclass
class RFF(AbstractKernel):
    r"""Computes an approximation of the kernel using Random Fourier Features.

    All stationary kernels are equivalent to the Fourier transform of a probability
    distribution. We call the corresponding distribution the spectral density. Using
    a finite number of basis functions, we can compute the spectral density using a
    Monte-Carlo approximation. This is done by sampling from the spectral density and
    computing the Fourier transform of the samples. The kernel is then approximated by
    the inner product of the Fourier transform of the samples with the Fourier
    transform of the data.

    The key reference for this implementation is the following papers:
    - 'Random Features for Large-Scale Kernel Machines' by Rahimi and Recht (2008).
    - 'On the Error of Random Fourier Features' by Sutherland and Schneider (2015).
    """

    base_kernel: AbstractKernel = None
    num_basis_fns: int = static_field(50)
    frequencies: Float[Array, "M 1"] = param_field(None, bijector=tfb.Identity())
    compute_engine: BasisFunctionComputation = static_field(
        BasisFunctionComputation(), repr=False
    )
    key: KeyArray = static_field(PRNGKey(123))

    def __post_init__(self) -> None:
        r"""Post-initialisation function.

        This function is called after the initialisation of the kernel. It is used to
        set the computation engine to be the basis function computation engine.
        """
        self._check_valid_base_kernel(self.base_kernel)

        if self.frequencies is None:
            n_dims = self.base_kernel.ndims
            self.frequencies = self.base_kernel.spectral_density.sample(
                seed=self.key, sample_shape=(self.num_basis_fns, n_dims)
            )
        self.name = f"{self.base_kernel.name} (RFF)"

    def __call__(self, x: Array, y: Array) -> Array:
        """Superfluous for RFFs."""

    def _check_valid_base_kernel(self, kernel: AbstractKernel):
        r"""Verify that the base kernel is valid for RFF approximation.

        Args:
            kernel (AbstractKernel): The kernel to be checked.
        """
        if kernel is None:
            raise ValueError("Base kernel must be specified.")
        error_msg = """
        Base kernel must have a spectral density. Currently, only Matérn
        and RBF kernels have implemented spectral densities.
        """
        if kernel.spectral_density is None:
            raise ValueError(error_msg)

    def compute_features(self, x: Float[Array, "N D"]) -> Float[Array, "N L"]:
        r"""Compute the features for the inputs.

        Args:
            x: A $`N \times D`$ array of inputs.

        Returns
        -------
            Float[Array, "N L"]: A $`N \times L`$ array of features where $`L = 2M`$.
        """
        return self.compute_engine.compute_features(self, x)



# Copyright 2022 The JaxLinOp Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from dataclasses import dataclass

from beartype.typing import Any
import jax.numpy as jnp
from jaxtyping import Float

from gpjax.linops.constant_diagonal_linear_operator import (
    ConstantDiagonalLinearOperator,
)
from gpjax.linops.utils import default_dtype
from gpjax.typing import (
    Array,
    ScalarFloat,
)


def _check_size(size: Any) -> None:
    """Check that size is an integer."""
    if not isinstance(size, int):
        raise ValueError(f"`size` must be an integer, but `size = {size}`.")


@dataclass
class IdentityLinearOperator(ConstantDiagonalLinearOperator):
    """Identity linear operator."""

    def __init__(self, size: int, dtype: jnp.dtype = None) -> None:
        """Identity matrix.

        Args:
            size (int): Size of the identity matrix.
        """
        _check_size(size)

        if dtype is None:
            dtype = default_dtype()

        self.value = jnp.array([1.0], dtype=dtype)
        self.size = size
        self.shape = (size, size)
        self.dtype = dtype

    def __matmul__(self, other: Float[Array, "N M"]) -> Float[Array, "N M"]:
        """Matrix multiplication.

        Args:
            other (Float[Array, "N M"]): Matrix to multiply with.

        Returns
        -------
            Float[Array, "N M"]: Result of matrix multiplication.
        """
        return other

    def to_root(self) -> "IdentityLinearOperator":
        """
        Lower triangular.

        Returns
        -------
            Float[Array, "N N"]: Lower triangular matrix.
        """
        return self

    def log_det(self) -> ScalarFloat:
        """Log determinant.

        Returns
        -------
            ScalarFloat: Log determinant of the covariance matrix.
        """
        return jnp.array(0.0)

    def inverse(self) -> "IdentityLinearOperator":
        """Inverse of the covariance operator.

        Returns
        -------
            DiagonalLinearOperator: Inverse of the covariance operator.
        """
        return self

    def solve(self, rhs: Float[Array, "... M"]) -> Float[Array, "... M"]:
        """Solve linear system.

        Args:
            rhs (Float[Array, "N M"]): Right hand side of the linear system.

        Returns
        -------
            Float[Array, "N M"]: Solution of the linear system.
        """
        # TODO: Check shapes.

        return rhs

    @classmethod
    def from_root(cls, root: "IdentityLinearOperator") -> "IdentityLinearOperator":
        """Construct from root.

        Args:
            root (IdentityLinearOperator): Root of the covariance operator.

        Returns
        -------
            IdentityLinearOperator: Covariance operator.
        """
        return root

    @classmethod
    def from_dense(cls, dense: Float[Array, "N N"]) -> "IdentityLinearOperator":
        return IdentityLinearOperator(dense.shape[0])


__all__ = [
    "IdentityLinearOperator",
]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from dataclasses import dataclass

import jax.numpy as jnp
from jaxtyping import Float
import tensorflow_probability.substrates.jax.bijectors as tfb

from gpjax.base import param_field
from gpjax.kernels.base import AbstractKernel
from gpjax.typing import (
    Array,
    ScalarFloat,
)


@dataclass
class Linear(AbstractKernel):
    r"""The linear kernel."""

    variance: ScalarFloat = param_field(jnp.array(1.0), bijector=tfb.Softplus())
    name: str = "Linear"

    def __call__(
        self,
        x: Float[Array, " D"],
        y: Float[Array, " D"],
    ) -> ScalarFloat:
        r"""Compute the linear kernel between a pair of arrays.

        For a pair of inputs $`x, y \in \mathbb{R}^{D}`$, let's evaluate the linear
        kernel $`k(x, y)=\sigma^2 x^{\top}y`$ where $`\sigma^\in \mathbb{R}_{>0}`$ is the
        kernel's variance parameter.

        Args:
            x (Float[Array, " D"]): The left hand input of the kernel function.
            y (Float[Array, " D"]): The right hand input of the kernel function.

        Returns
        -------
            ScalarFloat: The evaluated kernel function $`k(x, y)`$ at the supplied inputs.
        """
        x = self.slice_input(x)
        y = self.slice_input(y)
        K = self.variance * jnp.matmul(x.T, y)
        return K.squeeze()



# Copyright 2022 The JaxLinOp Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import jax.numpy as jnp
import jax.scipy as jsp
from jaxtyping import Float

from gpjax.linops.dense_linear_operator import DenseLinearOperator
from gpjax.linops.linear_operator import LinearOperator
from gpjax.typing import Array


class LowerTriangularLinearOperator(DenseLinearOperator):
    """Current implementation of the following methods is inefficient.
    We assume a dense matrix representation of the operator. But take advantage of the solve structure.
    """

    @property
    def T(self) -> "UpperTriangularLinearOperator":
        return UpperTriangularLinearOperator(matrix=self.matrix.T)

    def to_root(self) -> LinearOperator:
        raise ValueError("Matrix is not positive semi-definite.")

    def inverse(self) -> DenseLinearOperator:
        matrix = self.solve(jnp.eye(self.size))
        return DenseLinearOperator(matrix)

    def solve(self, rhs: Float[Array, "... M"]) -> Float[Array, "... M"]:
        return jsp.linalg.solve_triangular(self.to_dense(), rhs, lower=True)

    @classmethod
    def from_root(cls, root: LinearOperator) -> None:
        raise ValueError("LowerTriangularLinearOperator does not have a root.")

    @classmethod
    def from_dense(cls, dense: Float[Array, "N N"]) -> "LowerTriangularLinearOperator":
        return LowerTriangularLinearOperator(matrix=dense)


class UpperTriangularLinearOperator(DenseLinearOperator):
    """Current implementation of the following methods is inefficient.
    We assume a dense matrix representation of the operator. But take advantage of the solve structure.
    """

    @property
    def T(self) -> LowerTriangularLinearOperator:
        return LowerTriangularLinearOperator(matrix=self.matrix.T)

    def to_root(self) -> LinearOperator:
        raise ValueError("Matrix is not positive semi-definite.")

    def inverse(self) -> DenseLinearOperator:
        matrix = self.solve(jnp.eye(self.size))
        return DenseLinearOperator(matrix)

    def solve(self, rhs: Float[Array, "... M"]) -> Float[Array, "... M"]:
        return jsp.linalg.solve_triangular(self.to_dense(), rhs, lower=False)

    @classmethod
    def from_root(cls, root: LinearOperator) -> None:
        raise ValueError("LowerTriangularLinearOperator does not have a root.")

    @classmethod
    def from_dense(cls, dense: Float[Array, "N N"]) -> "UpperTriangularLinearOperator":
        return UpperTriangularLinearOperator(matrix=dense)


__all__ = ["LowerTriangularLinearOperator", "UpperTriangularLinearOperator"]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from dataclasses import dataclass

import jax.numpy as jnp
from jaxtyping import Float
import tensorflow_probability.substrates.jax.bijectors as tfb

from gpjax.base import (
    param_field,
    static_field,
)
from gpjax.kernels.base import AbstractKernel
from gpjax.typing import (
    Array,
    ScalarFloat,
    ScalarInt,
)


@dataclass
class Polynomial(AbstractKernel):
    """The Polynomial kernel with variable degree."""

    degree: ScalarInt = static_field(2)
    shift: ScalarFloat = param_field(jnp.array(1.0), bijector=tfb.Softplus())
    variance: ScalarFloat = param_field(jnp.array(1.0), bijector=tfb.Softplus())

    def __post_init__(self):
        self.name = f"Polynomial (degree {self.degree})"

    def __call__(self, x: Float[Array, " D"], y: Float[Array, " D"]) -> ScalarFloat:
        r"""Compute the polynomial kernel of degree $`d`$ between a pair of arrays.

        For a pair of inputs $`x, y \in \mathbb{R}^{D}`$, let's evaluate the polynomial
        kernel $`k(x, y)=\left( \alpha + \sigma^2 x y\right)^{d}`$ where
        $`\sigma^\in \mathbb{R}_{>0}`$ is the kernel's variance parameter, shift
        parameter $`\alpha`$ and integer degree $`d`$.

        Args:
            x (Float[Array, " D"]): The left hand argument of the kernel function's
                call.
            y (Float[Array, " D"]): The right hand argument of the kernel function's
                call

        Returns
        -------
            ScalarFloat: The value of $`k(x, y)`$.
        """
        x = self.slice_input(x)
        y = self.slice_input(y)
        K = jnp.power(self.shift + self.variance * jnp.dot(x, y), self.degree)
        return K.squeeze()



from dataclasses import dataclass
import typing as tp

import jax.numpy as jnp
from jaxtyping import Float

from gpjax.kernels.computations.base import AbstractKernelComputation
from gpjax.linops import DenseLinearOperator
from gpjax.typing import Array

Kernel = tp.TypeVar("Kernel", bound="gpjax.kernels.base.AbstractKernel")  # noqa: F821


@dataclass
class BasisFunctionComputation(AbstractKernelComputation):
    r"""Compute engine class for finite basis function approximations to a kernel."""

    def cross_covariance(
        self, kernel: Kernel, x: Float[Array, "N D"], y: Float[Array, "M D"]
    ) -> Float[Array, "N M"]:
        r"""Compute an approximate cross-covariance matrix.

        For a pair of inputs, compute the cross covariance matrix between the inputs.

        Args:
            kernel (Kernel): the kernel function.
            x: (Float[Array, "N D"]): A $`N \times D`$ array of inputs.
            y: (Float[Array, "M D"]): A $`M \times D`$ array of inputs.

        Returns:
            Float[Array, "N M"]: A $N \times M$ array of cross-covariances.
        """
        z1 = self.compute_features(kernel, x)
        z2 = self.compute_features(kernel, y)
        return self.scaling(kernel) * jnp.matmul(z1, z2.T)

    def gram(self, kernel: Kernel, inputs: Float[Array, "N D"]) -> DenseLinearOperator:
        r"""Compute an approximate Gram matrix.

        For the Gram matrix, we can save computations by computing only one matrix
        multiplication between the inputs and the scaled frequencies.

        Args:
            kernel (Kernel): the kernel function.
            inputs (Float[Array, "N D"]): A $`N x D`$ array of inputs.

        Returns:
            DenseLinearOperator: A dense linear operator representing the
                $`N \times N`$ Gram matrix.
        """
        z1 = self.compute_features(kernel, inputs)
        return DenseLinearOperator(self.scaling(kernel) * jnp.matmul(z1, z1.T))

    def compute_features(
        self, kernel: Kernel, x: Float[Array, "N D"]
    ) -> Float[Array, "N L"]:
        r"""Compute the features for the inputs.

        Args:
            kernel (Kernel): the kernel function.
            x (Float[Array, "N D"]): A $`N \times D`$ array of inputs.

        Returns
        -------
            Float[Array, "N L"]: A $`N \times L`$ array of features where $`L = 2M`$.
        """
        frequencies = kernel.frequencies
        scaling_factor = kernel.base_kernel.lengthscale
        z = jnp.matmul(x, (frequencies / scaling_factor).T)
        z = jnp.concatenate([jnp.cos(z), jnp.sin(z)], axis=-1)
        return z

    def scaling(self, kernel: Kernel):
        r"""Compute the scaling factor for the covariance matrix.

        Args:
            kernel (Kernel): the kernel function.

        Returns
        -------
            Float[Array, ""]: A scalar array.
        """
        return kernel.base_kernel.variance / kernel.num_basis_fns



# Copyright 2022 The JaxLinOp Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from dataclasses import dataclass

from beartype.typing import Union
import jax.numpy as jnp
from jaxtyping import Float

from gpjax.linops.linear_operator import LinearOperator
from gpjax.linops.utils import to_linear_operator
from gpjax.typing import (
    Array,
    ScalarFloat,
    VecNOrMatNM,
)


def _check_matrix(matrix: Array) -> None:
    if matrix.ndim != 2:
        raise ValueError(
            "The `matrix` must have at two dimensions, but "
            f"`scale.shape = {matrix.shape}`."
        )

    if matrix.shape[-1] != matrix.shape[-2]:
        raise ValueError(
            f"The `matrix` must be a square matrix, but `scale.shape = {matrix.shape}`."
        )


@dataclass
class DenseLinearOperator(LinearOperator):
    """Dense covariance operator."""

    matrix: Float[Array, "N N"]

    def __init__(
        self, matrix: Float[Array, "N N"], dtype: jnp.dtype = None, shape: tuple = None
    ) -> None:
        """Initialize the covariance operator.

        Args:
            matrix (Float[Array, "N N"]): Dense matrix.
        """
        _check_matrix(matrix)

        if dtype is not None:
            matrix = matrix.astype(dtype)

        self.matrix = matrix
        self.shape = matrix.shape
        self.dtype = matrix.dtype

    def __add__(
        self, other: Union[LinearOperator, Float[Array, "N N"]]
    ) -> LinearOperator:
        """Add diagonal to another linear operator.

        Args:
            other (Union[LinearOperator, Float[Array, "N N"]]): Other linear operator. Dimension of both operators must match. If the other linear operator is not a DiagonalLinearOperator, dense matrix addition is used.

        Returns
        -------
            LinearOperator: linear operator plus the diagonal linear operator.
        """
        from gpjax.linops.diagonal_linear_operator import DiagonalLinearOperator
        from gpjax.linops.zero_linear_operator import ZeroLinearOperator

        other = to_linear_operator(other)

        if isinstance(other, DiagonalLinearOperator):
            return self._add_diagonal(other)

        elif isinstance(other, DenseLinearOperator):
            return DenseLinearOperator(matrix=self.matrix + other.matrix)

        elif isinstance(other, ZeroLinearOperator):
            return self

        else:
            raise NotImplementedError

    def __mul__(self, other: ScalarFloat) -> LinearOperator:
        """Multiply covariance operator by scalar.

        Args:
            other (LinearOperator): Scalar.

        Returns
        -------
            LinearOperator: Covariance operator multiplied by a scalar.
        """
        return DenseLinearOperator(matrix=self.matrix * other)

    def _add_diagonal(
        self,
        other: "gpjax.linops.diagonal_linear_operator.DiagonalLinearOperator",  # noqa: F821
    ) -> LinearOperator:
        """Add diagonal to the covariance operator,  useful for computing, Kxx + Io².

        Args:
            other (DiagonalLinearOperator): Diagonal covariance operator to add to the covariance operator.

        Returns
        -------
            LinearOperator: Sum of the two covariance operators.
        """
        dim = self.shape[0]
        diag_indices = jnp.diag_indices(dim)
        new_matrix = self.matrix.at[diag_indices].add(other.diagonal())

        return DenseLinearOperator(matrix=new_matrix)

    def diagonal(self) -> Float[Array, " N"]:
        """
        Diagonal of the covariance operator.

        Returns
        -------
            Float[Array, " N"]: The diagonal of the covariance operator.
        """
        return jnp.diag(self.matrix)

    def __matmul__(self, other: VecNOrMatNM) -> VecNOrMatNM:
        """Matrix multiplication.

        Args:
            other (Float[Array, "N M"]): Matrix to multiply with.

        Returns
        -------
            Float[Array, "N M"]: Result of matrix multiplication.
        """
        return jnp.matmul(self.matrix, other)

    def to_dense(self) -> Float[Array, "N N"]:
        """Construct dense Covariance matrix from the covariance operator.

        Returns
        -------
            Float[Array, "N N"]: Dense covariance matrix.
        """
        return self.matrix

    @classmethod
    def from_dense(cls, matrix: Float[Array, "N N"]) -> "DenseLinearOperator":
        """Construct covariance operator from dense covariance matrix.

        Args:
            matrix (Float[Array, "N N"]): Dense covariance matrix.

        Returns
        -------
            DenseLinearOperator: Covariance operator.
        """
        return DenseLinearOperator(matrix=matrix)

    @classmethod
    def from_root(cls, root: LinearOperator) -> "DenseLinearOperator":
        """Construct covariance operator from the root of the covariance matrix.

        Args:
            root (Float[Array, "N N"]): Root of the covariance matrix.

        Returns
        -------
            DenseLinearOperator: Covariance operator.
        """
        return DenseFromRootLinearOperator(root=root)


class DenseFromRootLinearOperator(DenseLinearOperator):
    root: LinearOperator

    def __init__(self, root: LinearOperator):
        """Initialize the covariance operator."""
        self.root = root
        self.shape = root.shape
        self.dtype = root.dtype

    def to_root(self) -> LinearOperator:
        return self.root

    @property
    def matrix(self) -> Float[Array, "N N"]:
        dense_root = self.root.to_dense()
        return dense_root @ dense_root.T


__all__ = [
    "DenseLinearOperator",
]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from dataclasses import dataclass

import beartype.typing as tp
import jax.numpy as jnp
from jaxtyping import (
    Float,
    Num,
)

from gpjax.kernels.computations.base import AbstractKernelComputation
from gpjax.typing import Array

Kernel = tp.TypeVar("Kernel", bound="gpjax.kernels.base.AbstractKernel")  # noqa: F821


@dataclass
class EigenKernelComputation(AbstractKernelComputation):
    r"""Eigen kernel computation class. Kernels who operate on an
    eigen-decomposed structure should use this computation object.
    """

    def cross_covariance(
        self, kernel: Kernel, x: Num[Array, "N D"], y: Num[Array, "M D"]
    ) -> Float[Array, "N M"]:
        r"""Compute the cross-covariance matrix.

        For an $`N\times D`$ and $`M\times D`$ pair of matrices, evaluate the $`N \times M`$
        cross-covariance matrix.

        Args:
            kernel (Kernel): the kernel function.
            x (Float[Array,"N D"]): The input matrix.
            y (Float[Array,"M D"]): The input matrix.

        Returns:
            _type_: _description_
        """
        # Transform the eigenvalues of the graph Laplacian according to the
        # RBF kernel's SPDE form.
        S = jnp.power(
            kernel.eigenvalues
            + 2 * kernel.smoothness / kernel.lengthscale / kernel.lengthscale,
            -kernel.smoothness,
        )
        S = jnp.multiply(S, kernel.num_vertex / jnp.sum(S))
        # Scale the transform eigenvalues by the kernel variance
        S = jnp.multiply(S, kernel.variance)
        return kernel(x, y, S=S)



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc
from dataclasses import dataclass
from functools import partial

from beartype.typing import (
    Callable,
    List,
    Optional,
    Union,
)
import jax.numpy as jnp
from jaxtyping import (
    Float,
    Num,
)
import tensorflow_probability.substrates.jax.distributions as tfd

from gpjax.base import (
    Module,
    param_field,
    static_field,
)
from gpjax.kernels.computations import (
    AbstractKernelComputation,
    DenseKernelComputation,
)
from gpjax.typing import (
    Array,
    ScalarFloat,
)


@dataclass
class AbstractKernel(Module):
    r"""Base kernel class."""

    compute_engine: AbstractKernelComputation = static_field(DenseKernelComputation())
    active_dims: Optional[List[int]] = static_field(None)
    name: str = static_field("AbstractKernel")

    @property
    def ndims(self):
        return 1 if not self.active_dims else len(self.active_dims)

    def cross_covariance(self, x: Num[Array, "N D"], y: Num[Array, "M D"]):
        return self.compute_engine.cross_covariance(self, x, y)

    def gram(self, x: Num[Array, "N D"]):
        return self.compute_engine.gram(self, x)

    def slice_input(self, x: Float[Array, "... D"]) -> Float[Array, "... Q"]:
        r"""Slice out the relevant columns of the input matrix.

        Select the relevant columns of the supplied matrix to be used within the
        kernel's evaluation.

        Args:
            x (Float[Array, "... D"]): The matrix or vector that is to be sliced.

        Returns
        -------
            Float[Array, "... Q"]: A sliced form of the input matrix.
        """
        return x[..., self.active_dims] if self.active_dims is not None else x

    @abc.abstractmethod
    def __call__(
        self,
        x: Float[Array, " D"],
        y: Float[Array, " D"],
    ) -> ScalarFloat:
        r"""Evaluate the kernel on a pair of inputs.

        Args:
            x (Float[Array, " D"]): The left hand input of the kernel function.
            y (Float[Array, " D"]): The right hand input of the kernel function.

        Returns
        -------
            ScalarFloat: The evaluated kernel function at the supplied inputs.
        """
        raise NotImplementedError

    def __add__(self, other: Union["AbstractKernel", ScalarFloat]) -> "AbstractKernel":
        r"""Add two kernels together.
        Args:
            other (AbstractKernel): The kernel to be added to the current kernel.

        Returns
        -------
            AbstractKernel: A new kernel that is the sum of the two kernels.
        """
        if isinstance(other, AbstractKernel):
            return SumKernel(kernels=[self, other])
        else:
            return SumKernel(kernels=[self, Constant(other)])

    def __radd__(self, other: Union["AbstractKernel", ScalarFloat]) -> "AbstractKernel":
        r"""Add two kernels together.
        Args:
            other (AbstractKernel): The kernel to be added to the current kernel.

        Returns
        -------
            AbstractKernel: A new kernel that is the sum of the two kernels.
        """
        return self.__add__(other)

    def __mul__(self, other: Union["AbstractKernel", ScalarFloat]) -> "AbstractKernel":
        r"""Multiply two kernels together.

        Args:
            other (AbstractKernel): The kernel to be multiplied with the current kernel.

        Returns
        -------
            AbstractKernel: A new kernel that is the product of the two kernels.
        """
        if isinstance(other, AbstractKernel):
            return ProductKernel(kernels=[self, other])
        else:
            return ProductKernel(kernels=[self, Constant(other)])

    @property
    def spectral_density(self) -> Optional[tfd.Distribution]:
        return None


@dataclass
class Constant(AbstractKernel):
    r"""
    A constant kernel. This kernel evaluates to a constant for all inputs.
    The scalar value itself can be treated as a model hyperparameter and learned during training.
    """

    constant: ScalarFloat = param_field(jnp.array(0.0))

    def __call__(self, x: Float[Array, " D"], y: Float[Array, " D"]) -> ScalarFloat:
        r"""Evaluate the kernel on a pair of inputs.

        Args:
            x (Float[Array, " D"]): The left hand input of the kernel function.
            y (Float[Array, " D"]): The right hand input of the kernel function.

        Returns
        -------
            ScalarFloat: The evaluated kernel function at the supplied inputs.
        """
        return self.constant.squeeze()


@dataclass
class CombinationKernel(AbstractKernel):
    r"""A base class for products or sums of MeanFunctions."""

    kernels: List[AbstractKernel] = None
    operator: Callable = static_field(None)

    def __post_init__(self):
        # Add kernels to a list, flattening out instances of this class therein, as in GPFlow kernels.
        kernels_list: List[AbstractKernel] = []

        for kernel in self.kernels:
            if not isinstance(kernel, AbstractKernel):
                raise TypeError("can only combine Kernel instances")  # pragma: no cover

            if isinstance(kernel, self.__class__):
                kernels_list.extend(kernel.kernels)
            else:
                kernels_list.append(kernel)

        self.kernels = kernels_list

    def __call__(
        self,
        x: Float[Array, " D"],
        y: Float[Array, " D"],
    ) -> ScalarFloat:
        r"""Evaluate the kernel on a pair of inputs.

        Args:
            x (Float[Array, " D"]): The left hand input of the kernel function.
            y (Float[Array, " D"]): The right hand input of the kernel function.

        Returns
        -------
            ScalarFloat: The evaluated kernel function at the supplied inputs.
        """
        return self.operator(jnp.stack([k(x, y) for k in self.kernels]))


SumKernel = partial(CombinationKernel, operator=jnp.sum)
ProductKernel = partial(CombinationKernel, operator=jnp.prod)



# Copyright 2022 The JaxLinOp Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from dataclasses import dataclass

from beartype.typing import (
    Any,
    Union,
)
import jax.numpy as jnp
from jaxtyping import Float

from gpjax.linops.dense_linear_operator import DenseLinearOperator
from gpjax.linops.linear_operator import LinearOperator
from gpjax.linops.utils import to_linear_operator
from gpjax.typing import (
    Array,
    ScalarFloat,
    VecNOrMatNM,
)


def _check_diag(diag: Any) -> None:
    """Check if the diagonal is a vector."""
    if diag.ndim != 1:
        raise ValueError(
            "The `matrix` must be a one dimension vector, but "
            f"`diag.shape = {diag.shape}`."
        )


@dataclass
class DiagonalLinearOperator(LinearOperator):
    """Diagonal covariance operator."""

    diag: Float[Array, " N"]

    def __init__(self, diag: Float[Array, " N"], dtype: jnp.dtype = None) -> None:
        """Initialize the covariance operator.

        Args:
            diag (Float[Array, " N"]): Diagonal of the covariance operator.
        """
        _check_diag(diag)

        if dtype is not None:
            diag = diag.astype(dtype)

        dim = diag.shape[0]
        self.diag = diag
        self.shape = (dim, dim)
        self.dtype = diag.dtype

    def diagonal(self) -> Float[Array, " N"]:
        """Diagonal of the covariance operator.

        Returns
        -------
            Float[Array, " N"]: Diagonal of the covariance operator.
        """
        return self.diag

    def __add__(
        self, other: Union[LinearOperator, Float[Array, "N N"]]
    ) -> LinearOperator:
        """Add diagonal to another linear operator.

        Args:
            other (Union[LinearOperator, Float[Array, "N N"]]): Other linear operator. Dimension of both operators must match. If the other linear operator is not a DiagonalLinearOperator, dense matrix addition is used.

        Returns
        -------
            LinearOperator: linear operator plus the diagonal linear operator.
        """
        from gpjax.linops.zero_linear_operator import ZeroLinearOperator

        other = to_linear_operator(other)

        if isinstance(other, DiagonalLinearOperator):
            return DiagonalLinearOperator(diag=self.diagonal() + other.diagonal())

        elif isinstance(other, DenseLinearOperator):
            return other._add_diagonal(self)

        elif isinstance(other, ZeroLinearOperator):
            return self

        else:
            raise NotImplementedError

    def __mul__(self, other: ScalarFloat) -> LinearOperator:
        """Multiply covariance operator by scalar.

        Args:
            other (LinearOperator): Scalar.

        Returns
        -------
            LinearOperator: Covariance operator multiplied by a scalar.
        """
        return DiagonalLinearOperator(diag=self.diagonal() * other)

    def _add_diagonal(self, other: "DiagonalLinearOperator") -> LinearOperator:
        """Add diagonal to the covariance operator,  useful for computing, Kxx + Io².

        Args:
            other (DiagonalLinearOperator): Diagonal covariance operator to add to the covariance operator.

        Returns
        -------
            LinearOperator: Covariance operator with the diagonal added.
        """
        return DiagonalLinearOperator(diag=self.diagonal() + other.diagonal())

    def to_dense(self) -> Float[Array, "N N"]:
        """Construct dense Covariance matrix from the covariance operator.

        Returns
        -------
            Float[Array, "N N"]: Dense covariance matrix.
        """
        return jnp.diag(self.diagonal())

    def __matmul__(self, other: VecNOrMatNM) -> VecNOrMatNM:
        """Matrix multiplication.

        Args:
            other (Float[Array, "N M"]): Matrix to multiply with.

        Returns
        -------
            Float[Array, "N M"]: Result of matrix multiplication.
        """
        diag = (
            self.diagonal() if other.ndim == 1 else jnp.expand_dims(self.diagonal(), -1)
        )

        return diag * other

    def to_root(self) -> "DiagonalLinearOperator":
        """
        Lower triangular.

        Returns
        -------
            Float[Array, "N N"]: Lower triangular matrix.
        """
        return DiagonalLinearOperator(diag=jnp.sqrt(self.diagonal()))

    def log_det(self) -> ScalarFloat:
        """Log determinant.

        Returns
        -------
            ScalarFloat: Log determinant of the covariance matrix.
        """
        return jnp.sum(jnp.log(self.diagonal()))

    def inverse(self) -> "DiagonalLinearOperator":
        """Inverse of the covariance operator.

        Returns
        -------
            DiagonalLinearOperator: Inverse of the covariance operator.
        """
        return DiagonalLinearOperator(diag=1.0 / self.diagonal())

    def solve(self, rhs: VecNOrMatNM) -> VecNOrMatNM:
        """Solve linear system.

        Args:
            rhs (Float[Array, "N M"]): Right hand side of the linear system.

        Returns
        -------
            Float[Array, "N M"]: Solution of the linear system.
        """
        return self.inverse() @ rhs

    @classmethod
    def from_root(cls, root: "DiagonalLinearOperator") -> "DiagonalLinearOperator":
        """Construct covariance operator from the lower triangular matrix.

        Returns
        -------
            DiagonalLinearOperator: Covariance operator.
        """
        return DiagonalFromRootLinearOperator(root=root)

    @classmethod
    def from_dense(cls, dense: Float[Array, "N N"]) -> "DiagonalLinearOperator":
        """Construct covariance operator from its dense matrix representation.

        Returns
        -------
            DiagonalLinearOperator: Covariance operator.
        """
        return DiagonalLinearOperator(diag=dense.diagonal())


class DiagonalFromRootLinearOperator(DiagonalLinearOperator):
    root: DiagonalLinearOperator

    def __init__(self, root: DiagonalLinearOperator):
        """Initialize the covariance operator."""
        if not isinstance(root, DiagonalLinearOperator):
            raise ValueError("root must be a DiagonalLinearOperator")

        self.root = root
        self.shape = root.shape
        self.dtype = root.dtype

    def to_root(self) -> LinearOperator:
        return self.root

    @property
    def diag(self) -> Float[Array, " N"]:
        return self.root.diagonal() ** 2

    def diagonal(self) -> Float[Array, " N"]:
        return self.root.diagonal() ** 2


__all__ = [
    "DiagonalLinearOperator",
]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


__all__ = ["param_field"]

import dataclasses

from beartype.typing import (
    Any,
    Mapping,
    Optional,
)
import tensorflow_probability.substrates.jax.bijectors as tfb


def param_field(  # noqa: PLR0913
    default: Any = dataclasses.MISSING,
    *,
    bijector: Optional[tfb.Bijector] = None,
    trainable: bool = True,
    default_factory: Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: Optional[bool] = None,
    compare: bool = True,
    metadata: Optional[Mapping[str, Any]] = None,
):
    metadata = {} if metadata is None else dict(metadata)

    if "bijector" in metadata:
        raise ValueError("Cannot use metadata with `bijector` already set.")

    if "trainable" in metadata:
        raise ValueError("Cannot use metadata with `trainable` already set.")

    if "pytree_node" in metadata:
        raise ValueError("Cannot use metadata with `pytree_node` already set.")
    if bijector is None:
        bijector = tfb.Identity()
    metadata["bijector"] = bijector
    metadata["trainable"] = trainable
    metadata["pytree_node"] = True

    if (
        default is not dataclasses.MISSING
        and default_factory is not dataclasses.MISSING
    ):
        raise ValueError("Cannot specify both default and default_factory.")

    if default is not dataclasses.MISSING:
        default_factory = lambda: default

    return dataclasses.field(
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )
